"""
daemon.py
===========

Contains the daemon functions for the registrar

"""

import structlog
from typing import Callable, Dict

import redis


logger = structlog.getLogger(__name__)


# callback declaration: callable with 'config', item and replace
QueueItemHandler = Callable[[dict, str, bool], None]


def run_daemon(
    config: dict,
    replace: bool,
    host: str,
    port: int,
    handlers: Dict[str, QueueItemHandler],
    progress_set: str,
    failure_set: str,
    success_set: str,
):
    """Run the registrar daemon, listening on a redis queue
        for items to be (de-)registered.

    Args:
        config (dict): configuration for the registrar
        replace (bool): replace the files
        host (str): redis host url
        port (int): redis port
        handlers (Dict[str, QueueItemCallback]): a queue mapping dictionary,
            mapping the queue names to their respective handling callables
        progress_set (str): name of set for items in progress
        failure_set (str): name of set for failed items
        success_set (str): name of set for successfully processed items

    Raises:
        Exception: Raised if path is already registered
    """

    # initialize the queue client
    client = redis.Redis(host=host, port=port, charset="utf-8", decode_responses=True)
    queue_names = list(handlers.keys())
    logger.debug(
        f"waiting on items on redis "
        f"queue{'s' if len(queue_names) > 1 else ''}",
        queues=queue_names,
    )

    while True:
        # fetch an item from the queue to be handled
        queue, value = client.brpop(queue_names)

        try:
            client.sadd(progress_set, value)
            handler = handlers[queue]
            handler(config, value, replace)
            client.sadd(success_set, value)
        except Exception as e:
            logger.exception(e)
            client.sadd(failure_set, value)
        finally:
            client.srem(progress_set, value)
